""" postprocess utilities"""
import jax.numpy as jnp
from numpyro.infer import Predictive
from jax.numpy import DeviceArray


class ppc:
    """Posterior or prior predictive"""

    def __init__(self, model=None, method=None, guide=None,
                 svi_params=None, trace=None,
                 x: DeviceArray = None, rng_key=None, num_ppc: int = 100):

        assert model != None, f"For ppc, model instance needs to pass as an argument"
        self.model = model
        assert method != None, f"For ppc, method=svi or method=mcmc needs to pass as an argument"
        self.method = method
        if self.method == 'svi':
            assert guide != None, f"For ppc of SVI, guide needs to pass as an argument"
            self.guide = guide
            assert svi_params != None, f"For ppc of SVI, svi_params needs to pass as an argument"
            self.svi_params = svi_params
        if self.method == 'mcmc':
            assert trace != None, f"For ppc of mcmc, trace needs to pass as an argument"
            self.trace = trace
        assert x != None, f"For ppc, the input features are needed to pass as an argument"
        self.x = x
        self.rng_key = rng_key
        self.num_ppc = num_ppc
        print(self.num_ppc)

    def summary(self, samples):
        site_stats = {}
        q = [2.5, 97.5]
        for k, v in samples.items():
            site_stats[k] = {
                "mean": jnp.mean(v, axis=0),
                "std": jnp.std(v, 0),
                "q": jnp.percentile(v, q, axis=0)
            }
        return site_stats

    def pred(self):

        if self.method == 'svi':
            predictive = Predictive(model=self.model, guide=self.guide,
                                    params=self.svi_params, num_samples=self.num_ppc)
            samples = predictive(self.rng_key, x=self.x)
            self.model_predictions = self.summary(samples)

            self.parameters_predictions = self.guide.sample_posterior(self.rng_key, self.svi_result.params,
                                                                      sample_shape=(self.num_ppc,))



from typing import Optional
# dill for saving model
import dill
import pandas as pd
# jax import
import jax.random as random
import jax.numpy as jnp
from jax.numpy import DeviceArray
# numpyro imports
import numpyro
import numpyro.distributions as dist
from . infer import fit


class ModelHandler:
    """
    Represents a numpyro thermodynamic model handler.

    The construction of the model is based on two dataframes: sate_df and energy_df.

    state_df:
    Must includes the 'States' and 'Activity' columns. It should include the
    name of the energy as we want to have in the model and the corresponding coefficients for
    each state. These coefficients define the functional form of the energy as a linear combination
    between them.
    Example (state_df): Here is the state_df for sorsteq model.

    | State    | Activity | G_R | G_C | G_I |
    | -------- | -------- | --- | --- | --- |
    | Empty    | 0        | 0   | 0   | 0   |
    | CRP      | 0        | 0   | 1   | 0   |
    | RNAP     | 1        | 1   | 0   | 0   |
    | RNAP+CRP | 1        | 1   | 1   | 1   |

    energy_df:
    This dataframe provide the `Type` of energy as well as
    starting `Start` and stopping `Stop` location that the molecule for that
    specific energy is binding.

    Example (energy_df): Here is the state_df for sorsteq model.
    G_C: CRP energy.
    G_R: RNAP energy.
    G_I: Interaction energy.

    | Energies     | Type     | Start | Stop |
    | ------------ | -------- | ----- | ---- |
    | G_C          | additive | 1     | 27   |
    | G_R          | additive | 34    | 75   |
    | G_I          | scalar   | NA    | NA   |

    Parameters
    ----------
    L: (int)
        Length of each training sequence. Must be ``>= 1``.

    C: (int)
        Length of the alphabet in the sequence.

    state_df: (pd.DataFrame)
            The state dataframe.

    energy_df: (pd.DataFrame)
            The energy dataframe.

    D_H: (int)
        Number of nodes in the nonlinearity maps latent phenotype to
        measurements. Default = 20.

    kT: (float)
        Boltzmann constant. Default = 0.582 (room temperature).

    ge_noise_model_type: (str)
            Specifies the type of noise model the user wants to infer.
            The possible choices allowed: ['Gaussian','Cauchy','SkewedT', 'Empirical']

    """

    def __init__(self, L: int, C: int,
                 state_df: pd.DataFrame,
                 energy_df: pd.DataFrame,
                 D_H: int = 20, kT: float = 0.582,
                 ge_noise_model_type: str = 'Gaussian'):

        # Assing the sequence length.
        self.L = L
        # Assing the alphabet length.
        self.C = C
        # Assign the state_df to the layer.
        self.state_df = state_df
        # Assign the energy_df to the layer.
        self.energy_df = energy_df
        # Assign the number of hidden nodes in the GE regression.
        # Make new column in the energy_df for stop-stop length
        self.energy_df['l_lc'] = self.energy_df['Stop'] - \
            self.energy_df['Start']
        self.D_H = D_H
        # Assign the Boltzmann constant.
        self.kT = kT
        # Assign the ge noise model type.
        self.ge_noise_model_type = ge_noise_model_type

    def nonlin(self, x):
        """
        Define nonlinear function mapping the latent phenotype to measurements.
        """
        return jnp.tanh(x)

    def model(self, x: DeviceArray = None, y: DeviceArray = None,
              batch_size: int = None) -> DeviceArray:
        """
        Numpyro model instance.
        """
        L = self.L
        C = self.C
        D_H = self.D_H
        kT = self.kT
        energy_df = self.energy_df
        state_df = self.state_df

        # Get list of energy names
        energy_list = energy_df['Energies'].values

        # Initialize the theta dictionary
        theta_dict = {}

        # Create priors dictionary for theta
        for eng_name in energy_list:
            # Find the corresponding row in the energy_df
            ix = energy_df[energy_df['Energies'] == eng_name]
            # Find the type of energy to assign the theta.
            # There are two options 1. additive 2. scalar.
            eng_type = ix['Type'].values

            # Additive parameters
            if eng_type == 'additive':
                # Create the theta_0 name: insert theta_0 to the begining of the energy names
                theta_0_name = f'theta_0_{eng_name}'
                # Prior on the theta_0
                theta_dict[theta_0_name] = numpyro.sample(
                    theta_0_name, dist.Normal(loc=0, scale=1))
                # Create the theta_lc name: insert theta_lc to the begining of the energy names
                theta_lc_name = f'theta_lc_{eng_name}'
                # Find the shape of theta_lc
                theta_lc_shape = int(ix['l_lc'].values)
                # Prior on the theta_lc
                theta_dict[theta_lc_name] = numpyro.sample(theta_lc_name, dist.Normal(loc=jnp.zeros((theta_lc_shape, C)),
                                                                                      scale=jnp.ones((theta_lc_shape, C))))
            # Scalar parameters
            if eng_type == 'scalar':
                theta_0_name = f'theta_0{eng_name}'
                # Prior on the theta_0
                theta_dict[theta_0_name] = numpyro.sample(
                    theta_0_name, dist.Normal(loc=0, scale=1))

        # Initialize the energy dictionary
        G_dict = {}

        for eng_name in energy_list:
            # Find the corresponding row in the energy_df
            ix = self.energy_df[self.energy_df['Energies'] == eng_name]
            # Find the type of energy which used to assign the theta dict.
            # There are two options 1. additive 2. scalar
            eng_type = ix['Type'].values

            # If there is additive G, reterive theta and calculate G values
            if eng_type == 'additive':
                # Get the name of theta_0 parameters
                theta_0_name = f'theta_0_{eng_name}'
                # Get the name of theta_lc parameters
                theta_lc_name = f'theta_lc_{eng_name}'
                # Find the starting position on the sequence
                start_idx = int(ix['Start'].values)
                # Find the stopping position on the sequence
                stop_idx = int(ix['Stop'].values)
                # Find the length of interest on the sequence
                l_lc = int(ix['l_lc'].values)
                # Reshape input to samples x length x characters
                x_eng = x[:, C * start_idx:C * stop_idx]
                x_lc = jnp.reshape(x_eng, [-1, l_lc, C])
                # Compute Delta G
                G_dict[eng_name] = numpyro.deterministic(f'G_{eng_name}',
                                                         theta_dict[theta_0_name] + jnp.einsum('ij,kij->k', theta_dict[theta_lc_name], x_lc))

            # If the energy type is scalar, the G=theta
            if eng_type == 'scalar':
                theta_0_name = f'theta_0_{eng_name}'
                G_dict[eng_name] = theta_dict[theta_0_name]

        # Compute partition function
        # Z: total partion function
        Z = 0.0
        # prob: total probability
        prob = 0.0

        # Loop over States in state_df
        for state in state_df['States']:
            # Compute total energy for each state
            total_eng_state = 0.0

            for eng_name in energy_list:
                # Find the index of energy in the state_df
                jx = state_df[state_df['States'] == state]
                # get the coefficient of the energy for each state from the state_df
                coef = float(jx[eng_name].values)
                # For each state partial partion function values is coef*G_dict[eng_name]
                total_eng_state = total_eng_state + coef * G_dict[eng_name]
            # parital partion function for each state
            Z_state = jnp.exp(-total_eng_state / kT)

            # Get the activity of each state
            activity = float(jx['Activity'].values)
            # probability of the state
            prob_state = activity * Z_state

            # Update total partition function
            Z = Z + Z_state
            # Update total probability
            prob = prob + prob_state

        # Return the latent variable
        phi = numpyro.deterministic('phi', prob / Z)
        phi = phi[..., jnp.newaxis]
        if y is not None:
            assert phi.shape == y.shape, f"phi has shape {phi.shape}, y has shape {y.shape}"

        # GE parameters
        a = numpyro.sample("a", dist.Normal(loc=0, scale=1))
        b = numpyro.sample("b", dist.Normal(
            jnp.zeros((D_H, 1)), jnp.ones((D_H, 1))))
        c = numpyro.sample("c", dist.Normal(
            jnp.zeros((D_H, 1)), jnp.ones((D_H, 1))))
        d = numpyro.sample("d", dist.Normal(
            jnp.zeros((D_H, )), jnp.ones((D_H, ))))

        # GE regression
        tmp = jnp.einsum('ij, kj->ki', c, phi)
        g = numpyro.deterministic(
            "g", a + jnp.einsum('ij, ki->kj', b, self.nonlin(tmp + d[None, :])))
        if y is not None:
            assert g.shape == y.shape, f"g has shape {g.shape}, y has shape {y.shape}"

        # noise = numpyro.sample("noise", dist.Gamma(3.0, 1.0))
        self.alpha, self.beta, noise = self.noise_model(
            self.ge_noise_model_type)
        sigma_obs = 1.0 / jnp.sqrt(noise)
        with numpyro.plate("data", x.shape[0], subsample_size=batch_size) as ind:
            if y is not None:
                batch_y = y[ind]
            else:
                batch_y = None
            batch_g = g[ind]
            return numpyro.sample("yhat", dist.Normal(
                batch_g, sigma_obs).to_event(1), obs=batch_y)

    def noise_model(self, ge_noise_model_type):
        """
        Define the Global Epistasis Noise Model.
        """

        # Check the ge_noise_model_type is in the list of implemented one.
        error_mess = f"ge_noise_model_type should be 'Gamma' "
        assert ge_noise_model_type in ['Gamma'], error_mess

        # Gamma noise model
        if ge_noise_model_type == 'Gamma':
            alpha = numpyro.sample('alpha', dist.Uniform(0.5, 5))
            beta = numpyro.sample('beta', dist.Uniform(0.5, 2))
            return alpha, beta, numpyro.sample("noise", dist.Gamma(alpha, beta))
    # use the fit class as the ModelHandler instance

    def fit(self, args, x: DeviceArray, y: DeviceArray, rng_key: Optional[int] = None):
        if rng_key is None:
            rng_key, rng_key_predict = random.split(random.PRNGKey(0))
        self.method = args.method
        if args.method == 'svi':
            self.guide, self.svi_results = fit(
                args=args, rng_key=rng_key, model=self.model).svi(x=x, y=y)
            return self.guide, self.svi_results
        if args.method == 'mcmc':
            self.trace = fit(args=args, rng_key=rng_key,
                             model=self.model).mcmc(x=x, y=y)
            return self.trace

    def save(self, filepath=None):
        # This is not working need change.
        self.filepath = filepath
        print(f'Saving model to {self.filepath}')
        output_dict = {}
        if self.method == 'svi':
            # output_dict['model'] = self.model
            output_dict['guide'] = self.guide
            output_dict['svi_params'] = self.svi_results.params
            with open(self.filepath, 'wb') as handle:
                dill.dump(output_dict, handle)
        if self.method == 'mcmc':
            output_dict['model'] = self.model
            output_dict['trace'] = self.trace
            with open(self.filepath, 'wb') as handle:
                dill.dump(output_dict, handle)



"""Utilities."""

from typing import Optional
import time
from xmlrpc.client import boolean
import numpy as np
import jax.numpy as jnp
import pandas as pd

# Define built-in alphabets
alphabet_dict = {
    "dna": np.array(["A", "C", "G", "T"]),
    "rna": np.array(["A", "C", "G", "U"]),
    "protein": np.array(
        [
            "A",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "K",
            "L",
            "M",
            "N",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "V",
            "W",
            "Y",
        ]
    ),
    "protein*": np.array(
        [
            "A",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "K",
            "L",
            "M",
            "N",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "V",
            "W",
            "Y",
            "*",
        ]
    ),
}


def x_to_ohe(x: np.ndarray, alphabet: dict, ravel_seqs=True):
    """
    Convert a sequence array to a one-hot encoded matrix.

    Parameters
    ----------
    x: (np.ndarray)
        (N,) array of input sequences, each of length L

    alphabet: (np.ndarray)
        (C,) array describing the alphabet sequences are drawn from.

    ravel_seqs: (bool)
        Whether to return an (N, L*C) array, as opposed to an (N, L, C) array.

    Returns
    -------
    x_ohe: (np.ndarray)
        Array of one-hot encoded sequences, stored as np.int8 values.
    """
    # Get dimensions
    L = len(x[0])
    N = len(x)
    C = len(alphabet)

    # Shape sequences as array of int8s
    x_arr = np.frombuffer(bytes("".join(x), "utf-8"),
                          np.int8, N * L).reshape([N, L])

    # Create alphabet as array of int8s
    alphabet_arr = np.frombuffer(bytes("".join(alphabet), "utf-8"), np.int8, C)

    # Compute (N,L,C) grid of one-hot encoded values
    x_nlc = (x_arr[:, :, np.newaxis] ==
             alphabet_arr[np.newaxis, np.newaxis, :]).astype(np.int8)

    # Ravel if requested
    if ravel_seqs:
        x_ohe = x_nlc.reshape([N, L * C])
    else:
        x_ohe = x_nlc

    return x_ohe


def x_to_stats(x, alphabet, weights=None, verbose=False):
    """
    Identify the consensus sequence from a sequence alignment.

    Parameters
    ----------
    x: (np.ndarray)
        List of sequences. Sequences must all be the same length.

    alphabet: (np.ndarray)
        Alphabet from which sequences are drawn.

    weights: (None, np.ndarray)
        Weights for each sequence. E.g., count values, or numerical y values.
        If None, a value of 1 will be assumed for each sequence.

    verbose: (bool)
        Whether to print computation time.

    Returns
    -------
    consensus_seq: (str)
        Consensus sequence.
    """
    # Start timer
    start_time = time.time()

    # Get alphabet
    alphabet = alphabet_dict[alphabet]

    # Check weights and set if not provided
    if weights is None:
        weights = np.ones(len(x))
    else:
        weights = weights.astype(float)
        check(
            len(weights) == len(x),
            f"len(weights)={len(weights)} does not match len(x)={len(x)}",
        )

    # Do one-hot encoding of sequences
    t = time.time()
    x_nlc = x_to_ohe(x, alphabet, ravel_seqs=False)
    N, L, C = x_nlc.shape

    # Dictionary to hold results
    stats = {}

    # Compute x_ohe
    stats["x_ohe_nonravel"] = x_nlc.astype(np.int8)
    stats["x_ohe"] = x_nlc.reshape([N, L * C]).astype(np.int8)

    # Multiply by weights
    x_nlc = x_nlc.astype(float) * weights[:, np.newaxis, np.newaxis]

    # Compute lc encoding of consensus sequence
    x_sum_lc = x_nlc.sum(axis=0)
    x_sum_lc = x_sum_lc.reshape([L, C])
    x_support_lc = x_sum_lc != 0

    # Set number of sequences
    stats["N"] = N

    # Set sequence length
    stats["L"] = L

    # Set number of characters
    stats["C"] = C

    # Set alphabet
    stats["alphabet"] = alphabet

    # Compute probability matrix
    p_lc = x_sum_lc / x_sum_lc.sum(axis=1)[:, np.newaxis]
    stats["probability_df"] = pd.DataFrame(
        index=range(L), columns=alphabet, data=p_lc)

    # Compute sparsity factor
    stats["sparsity_factor"] = (x_nlc != 0).sum().sum() / x_nlc.size

    # Compute the consensus sequence and corresponding matrix.
    # Adding noise prevents ties
    x_sum_lc += 1e-1 * np.random.rand(*x_sum_lc.shape)
    stats["consensus_seq"] = "".join(
        [alphabet[np.argmax(x_sum_lc[l, :])] for l in range(L)])

    # Compute mask dict
    missing_dict = {}
    for l in range(L):
        if any(~x_support_lc[l, :]):
            missing_dict[l] = "".join(alphabet[~x_support_lc[l, :]])
    stats["missing_char_dict"] = missing_dict

    # Provide feedback if requested
    duration_time = time.time() - start_time
    if verbose:
        print(f"\nStats computation time: {duration_time:.5f} sec.\n")

    return stats


def load_dataset(filename: Optional[str] = None,
                 alphabet: Optional[np.array] = 'protein',
                 verbose: Optional[boolean] = True):
    """
    Load dataset provided in data directory.
    Parameters
    ----------
    filename: (str)
        Path to the data set.
    Returns
    -------
    x: (jnp.array)
        One-hot encoded of the sequences.
    y: (jnp.array)
        Measurements.
    L: (int):
        sequence length.
    C: (int)
        alphabet length.

    """
    # Load the dataset
    data_df = pd.read_csv(filename)
    # Get and report sequence length
    L = len(data_df.loc[0, "x"])
    print(f"Sequence length: {L:d} amino acids")
    # protein alphabet length
    C = 20

    # Get the training dataset (x,y)
    # Convert the sequences to the one-hot encoed array

    x_stats = x_to_stats(data_df["x"], alphabet=alphabet, verbose=True)
    x = x_stats["x_ohe_nonravel"]
    # training data is in log2 format
    y = data_df["y"].values

    if verbose == True:
        # Preview dataset
        print(f"\n Dataset looks like:")
        print(data_df.head())
        print(f'\nDataset consists of {len(data_df)} sequence\n')
    return jnp.array(x), jnp.array(y).reshape(-1, 1), L, C



# standard import
import time
# numpyro imports
from numpyro.infer import SVI, autoguide
from numpyro.infer import init_to_sample
from numpyro.infer import Trace_ELBO
import numpyro.optim as optim
from numpyro.infer import NUTS, MCMC
# jax imports
from jax.numpy import DeviceArray
import jax.numpy as jnp


class fit():
    def __init__(self,
                 args,
                 rng_key,
                 model: DeviceArray = None):

        self.rng_key = rng_key
        self.model = model
        self.device = args.device

        error_samp = f"number of samples for mcmc/svi inference should be provided as args"
        # Assign number of samples for inference.
        assert args.num_samples != None, error_samp
        self.num_samples = args.num_samples

        # Assign parameters for the MCMC inference
        if args.method == 'mcmc':
            error_warm = f"number of warmup steps for mcmc inference should be provided as args"
            error_chain = f"number of chains for mcmc inference should be provided as args"

            # Assign number of warmup steps for MCMC inference.
            assert args.num_warmup != None, error_warm
            self.num_warmup = args.num_warmup

            # Assign number of chains for MCMC inference.
            assert args.num_chains != None, error_chain
            self.num_chains = args.num_chains

        # Assign parameters for the SVI inference.
        if args.method == 'svi':
            # Assign the learning rate
            learning_rate = getattr(args, 'learning_rate', None)
            if learning_rate == None:
                self.step_size = 1e-1
                print('\nDefault Learning rate = 1e-1 is used')
            else:
                self.step_size = learning_rate
            # Assign the batch size
            batch_size = getattr(args, 'batch_size', None)
            if batch_size == None:
                self.batch_size = 128
                print('\nDefault batch size = 128 is used')
            else:
                self.batch_size = batch_size

            # Assign Learning rate decay
            learning_decay = getattr(args, 'learning_decay', None)
            if learning_decay == None:
                self.learning_decay = None
            else:
                self.learning_decay = learning_decay

    def svi(self, x: DeviceArray, y: DeviceArray):
        """
        Stochastic Variational Inference.

        Parameters
        ----------
        x: (DeviceArray):
            Input features for training.
        y: (DeviceArray):
            Output (measurements) for training.

        Returns
        -------
        guide: (numpyro.infer.autoguide.AutoDelta):
            Numpyro Automatic Guide Generated.

        svi_results: (SVIRunResult)
            Stochastic Variational Inference Result.
        """

        print('\nTraining using Stochastic Variational Inference\n')
        start = time.time()

        # Assign Autodelta Autoguide.
        guide = autoguide.AutoDelta(
            self.model, init_loc_fn=init_to_sample)

        # Initial learning rate
        step_size = self.step_size
        init_lr = self.step_size

        # Reduce learning rate based on the rate in learning decay for
        # every 1000 epochs
        if self.learning_decay is not None:
            def step_size(i):
                return init_lr * self.learning_decay**jnp.floor(i / 1_000)

        # Define the optimizer
        optimizer = optim.RMSProp(step_size=step_size)

        # Loss function is the ELBO
        svi = SVI(self.model, guide, optimizer, loss=Trace_ELBO())

        svi_results = svi.run(
            rng_key=self.rng_key, num_steps=self.num_samples,
            x=x, y=y, batch_size=self.batch_size)

        print("\nVariational inference elapsed time:", time.time() - start)
        return guide, svi_results

    def mcmc(self, x, y):
        print('\nTraining using MCMC\n')
        start = time.time()
        # define kernel
        kernel = NUTS(model=self.model)
        # setup mcmc
        mcmc = MCMC(kernel,
                    num_warmup=self.num_warmup,
                    num_samples=self.num_samples,
                    num_chains=self.num_chains,
                    progress_bar=False if self.device is 'gpu' else True)
        # run mcmc inference
        mcmc.run(self.rng_key, x=x, y=y)
        print("\nMCMC elapsed time:", time.time() - start)
        return mcmc
